from . import BaseActor
from ltr.models.loss.target_regression import DenseIouPred
import torch

class TrSegActor(BaseActor):
    """Actor for jointly training cls_18, cls_72 branch and regression branch, except for regression optimizer"""
    def __init__(self, net, objective, device, output_sz, loss_weight=None):
        super().__init__(net, objective, device)
        self.loss_weight = loss_weight
        self.output_sz = output_sz

    def box_cxcywh_to_xyxy(self, x):
        x_c, y_c, w, h = x.unbind(-1)
        b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
             (x_c + 0.5 * w), (y_c + 0.5 * h)]
        return torch.stack(b, dim=-1)

    def box_xywh_to_xyxy(self, x):
        x1, y1, w, h = x.unbind(-1)
        b = [x1, y1, x1 + w, y1 + h]
        return torch.stack(b, dim=-1)

    def __call__(self, data):
        """
        args:
            data - The input data.
            gen_iou_map -
        returns:
            loss    - the training loss
            stats  -  dict containing detailed losses
        """
        # Run network
        # mask: (batch,1,256,256)  corner: (batch,4) --- (x0,y0,x1,y1)
        output_mask, output_corner = self.net(train_imgs=data['train_images'],
                                              test_imgs=data['test_images'],
                                              train_bb=data['train_anno'],
                                              train_mask=data['train_masks'])

        ### bbox loss loss ###
        gt_box = self.box_xywh_to_xyxy(data['test_anno'].clone().view(-1, 4))  # X1Y1WH --> x1y1x2y2
        bbox_loss = self.objective['bbox'](output_corner, gt_box) * self.loss_weight['bbox']

        ### Regression offline loss ###
        # print(data['test_anno'].shape)
        # print(output_coord.shape)
        # gt = self.box_xywh_to_xyxy(data['test_anno'].clone().view(-1, 4)) / self.output_sz # X1Y1WH --> x1y1x2y2
        # # output = output_coord.view(-1, 4)
        # output = self.box_cxcywh_to_xyxy(output_coord.view(-1, 4)) # XcYcWH --> x1y1x2y2
        # l1_loss = self.objective['l1'](output, gt) * self.loss_weight['l1']
        #
        # giou_loss, iou = self.objective['giou'](output, gt)
        # iou = iou.mean()
        # giou_loss = giou_loss * self.loss_weight['giou']

        vis_gt_mask = None
        vis_pred_mask = None
        vis_test_img = None
        ## mask loss
        mask_flag = (data['mask'] == 1)  # data[mask]是一个tensor,有batch个元素,其中有的是0有的是1. mask_flag是一个bool型tensor
        num_mask_sample = mask_flag.sum()
        if num_mask_sample > 0:
            mask_gt = data['test_masks'].squeeze(0)  # 测试帧的mask真值框 (batch,1,H,W)
            vis_gt_mask = mask_gt[mask_flag][0].squeeze()  # h, w
            vis_pred_mask = output_mask[mask_flag][0].squeeze()  # h, w
            vis_test_img = data['test_images'][0][mask_flag][0].unsqueeze(0)  # 1, c, h, w
            '''Compute loss for mask'''
            loss_mask = self.objective['mask'](output_mask[mask_flag], mask_gt[mask_flag])  # 只计算那些mask_flag等于1的样本的loss_mask
        else:
            loss_mask = torch.zeros((1,))
            vis_pred_mask = output_mask[0].squeeze()
            vis_test_img = data['test_images'][0][0].unsqueeze(0)

        mask_loss = loss_mask * self.loss_weight['mask']

        # Total loss
        loss =  mask_loss + bbox_loss

        # Log stats
        stats = {'Loss/total': loss.item(),
                 'Loss/box': bbox_loss.item(),
                 'Loss/mask': mask_loss.item()
                }

        return loss, stats, vis_test_img, vis_gt_mask, vis_pred_mask

