import math
import torch.nn as nn
from collections import OrderedDict
import ltr.models.backbone as backbones
from ltr import model_constructor
from typing import Optional, List
from torch import nn, Tensor
import torch
import os, cv2
import torch.nn.functional as F
from ltr.models.target_classifier.initializer import FilterPool
from ltr.models.trSeg.head import build_box_head
from ltr.models.trSeg.transformer_v2 import Transformer
import torch.distributed as dist
from ltr.models.trSeg.head import Mask_Head

class TrSeg(nn.Module):
    """The FCOTNet network.
    args:
        feature_extractor:  Backbone feature extractor network. Must return a dict of feature maps.
        classification_layer:  Name of the backbone feature layer to use for classification.
        pyramid_*: The blocks in Decoder.
        classifier_18:  Target classification module with feature size of 18.
        classifier_72:  Target classification module with feature size of 72.
        regressor_72:  Anchor-free regression module with feature size of 72.
        train_cls_72_and_reg_init: Whether clssifier_72 and regression initial model should be trained or not.
        train_reg_optimizer:  Whether regression optimizer should be trained or not.
        train_cls_18:  Whether classifier_18 should be trained or not. """

    def __init__(self, feature_extractor, classification_layer, transformer, mask_head, num_queries=1, hidden_dim=256):
        super().__init__()

        self.feature_extractor = feature_extractor
        self.transformer = transformer

        self.target_pool = FilterPool(filter_size=8, feature_stride=16, pool_square=False)
        # self.query_embed = nn.Embedding(num_queries, hidden_dim)  # object queries
        # self.pos_embed = nn.Embedding(5*5+18*18, hidden_dim)  # positional embedding
        self.proj = nn.Linear(1024, hidden_dim)
        self.proj2 = nn.Linear(hidden_dim, 64)
        # self.activation = nn.ReLU(inplace=True)

        self.mask_head = mask_head

        self.classification_layer = [classification_layer] if isinstance(classification_layer, str) else classification_layer
        self.output_layers = ['layer1', 'layer2', 'layer3']

    def forward(self, train_imgs, test_imgs, train_bb, *args, **kwargs):

        assert train_imgs.dim() == 5 and test_imgs.dim() == 5, 'Expect 5 dimensional inputs'

        _, n, _,_,_ = train_imgs.shape

        # Extract backbone features
        train_feat = self.extract_backbone_features(train_imgs.view(-1, *train_imgs.shape[-3:]),
                                                    layers=['conv1', 'layer1','layer2','layer3'])
        test_feat = self.extract_backbone_features(test_imgs.view(-1, *test_imgs.shape[-3:]),
                                                   layers=['conv1', 'layer1','layer2','layer3'])

        Lfeat_list = [feat for name, feat in test_feat.items() if name != 'layer3']

        train_feat = self.get_backbone_feat(train_feat) # n_images * B, C, Ht, Wt
        test_feat = self.get_backbone_feat(test_feat) # n_images * B, C, Hs, Ws
        # print("train_feat:{}".format(train_feat.shape))
        _, _, hs, ws = test_feat.size()
        target = self.generate_target(train_feat, train_bb)  # n_images * B, C, Ht, Wt
        # print(train_bb)

        search = self.proj(test_feat.flatten(start_dim=2).permute(2, 0, 1))  # HsWsxNxC
        target = self.proj(target.flatten(2).permute(2, 0, 1))

        queried_sear = self.proj2(search)
        # queried_sear = self.transformer(search, target)[0]  # HW, n, C  (C=64)
        # print("out_emm:{}".format(outputs_embed.shape))
        queried_sear = queried_sear.permute(1, 2, 0).view(n, -1, hs, ws)  # n, 64, 16, 16
        outputs_mask = self.mask_head(queried_sear, Lfeat_list)    # n, 1, 256, 256

        # outputs_coord = self.mask_to_head().view(-1, 4) # n, 4
        # print("out_coor:{}".format(outputs_coord.shape))

        return outputs_mask

    # def mask_to_head(self, mask):
    #     """
    #     mask: output mask (N, 1, 256, 256)
    #     """
    #     # Forward the class and box head
    #
    #
    #     return outputs_coord

    # def forward_box_head(self, hs):
    #     """
    #     hs: output embeddings (NB, C)
    #     """
    #     # Forward the class and box head
    #     hs1 = hs[:1, ...]
    #     hs2 = hs[1:, ...]
    #     hs2 = self.box_head2(hs2.permute(2, 1, 0)).permute(2, 1, 0)  # 1, nb, C
    #     mer_hs = torch.cat([hs1, hs2], dim=0)
    #     hs = self.merge_heads(mer_hs.permute(2, 1, 0)).permute(2, 1, 0)
    #     outputs_coord = self.box_head(hs).sigmoid()  # NB, 4
    #
    #     return outputs_coord

    def merge_target_search_query(self, search, target, query=None):
        """
        :param search: # n_images * B, C, Hs, Ws
        :param target: # n_images * B, C, Ht, Wt
        :param query: # 1, nb, C
        :return:
        """
        # search = search.flatten(start_dim=2).permute(2,0,1) # HWxNBxC
        # target = target.flatten(2).permute(2,0,1)

        if query is not None:
            src = torch.cat([search, target, query], dim=0)
        else:
            src = torch.cat([search, target], dim=0)
        return src

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos


    def generate_target(self, feat_train, bb):
        bb_pool = bb.clone()
        # print("bb: {}".format(bb))
        target = self.target_pool(feat_train, bb_pool)

        # target = target.view(-1, self.filter_dim, target.shape[-3] // self.filter_dim, target.shape[-2], target.shape[-1])
        return target

    def get_backbone_feat(self, backbone_feat):
        feat = OrderedDict({l: backbone_feat[l] for l in self.classification_layer})
        if len(self.classification_layer) == 1:
            return feat[self.classification_layer[0]]
        return feat

    def extract_classification_feat_18(self, backbone_feat):
        return self.classifier_18.extract_classification_feat(self.get_backbone_clf_feat(backbone_feat))

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = self.output_layers
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers=None):
        if layers is None:
            layers = self.bb_regressor_layer + ['classification']
        if 'classification' not in layers:
            return self.feature_extractor(im, layers)
        backbone_layers = sorted(list(set([l for l in layers + self.classification_layer if l != 'classification'])))
        all_feat = self.feature_extractor(im, backbone_layers)
        all_feat['classification'] = self.extract_classification_feat(all_feat)
        return OrderedDict({l: all_feat[l] for l in layers})

def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True

def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()

def is_main_process():
    return get_rank() == 0

@model_constructor
def transSeg(classification_layer='layer3', backbone_pretrained=True, freeze_bn=False,
                frozen_backbone_layers=()):
    # Backbone
    # print("main_proc: {}".format(is_main_process()))
    backbone_net = backbones.resnet50(pretrained=is_main_process(), frozen_layers=frozen_backbone_layers, freeze_bn=freeze_bn)

    transformer = Transformer(d_model=256, nhead=8, num_encoder_layers=6, num_decoder_layers=6)

    mask_head = Mask_Head()

    net = TrSeg(feature_extractor=backbone_net, classification_layer=classification_layer,
                transformer=transformer, mask_head=mask_head)
    return net